//********************************************************************
//  BookOfSFKH.java @version 1.03
//    Simple simulator for Star Frontiers: Knight Hawks
//    space combat game, to price-balance units.
//    Based on BookOfWar v1.23 template
//  Copyright (c) 2013 Daniel R. Collins. All rights reserved.
//  See the bottom of this file for any licensing information.
//********************************************************************

public class BookOfSFKH {

	//-----------------------------------------------------------------
	//  Constants & switches
	//-----------------------------------------------------------------

	// Type of simulation to run
	// (0 = uniform assess, 1 = uniform detail, 2 = combo assess, 3 = combo detail)
	static final int simSelect = 0;

	// Uniform random budget range
	final int uniBudgetMin = 10;
	final int uniBudgetMax = 30;

	// Uniform ships to fight in detail
	final String uniDetailType1 = "FF";
	final String uniDetailType2 = "CV";

	// Combination squads to fight in detail
	final int comboDetailSquad1 = 0;
	final int comboDetailSquad2 = 14;

	// Trials per matchup (10K is ~1% error)
	final int trialsPerMatchup = 10000;

	// Create partial ship from budget remainder
	final boolean partialShipFromRemainder = true;

	// Allow shooters to redirect fire after a target destroyed
	final boolean optimizeTargetSelection = true;

	// Percent of time forward-firing cannons usable on defense
	final int percentFFOnDefense = 0;

	// ICMs shot at any incoming assault rocket or torpedo
	final int ICMsUsedPerAttack = 2;

	//-----------------------------------------------------------------
	//  Fighter ship type (launched from CV)
	//-----------------------------------------------------------------
	Ship fighterShipType =
		new Ship("VF", "Fighter",           8, 5, 5, 0, 0,  0, 3, 0, 0,  0, 0,  1);

	//-----------------------------------------------------------------
	//  Basic ship types available
	//-----------------------------------------------------------------
	Ship[] basicShipType = {
		new Ship("VS", "Scout",            25, 5, 4, 1, 0,  0, 4, 0, 0,  0, 0,  2),
		new Ship("FF", "Frigate",          40, 3, 3, 1, 1,  4, 0, 2, 1,  4, 0,  3),
		new Ship("DD", "Destroyer",        50, 3, 2, 1, 1,  6, 0, 2, 2,  4, 0,  4),
		new Ship("CL", "Light Cruiser",    70, 2, 2, 2, 1,  7, 0, 4, 2,  6, 0,  5),
		new Ship("CA", "Heavy Cruiser",    80, 1, 1, 3, 1,  8, 0, 4, 1,  8, 0,  6),
		new Ship("BB", "Battleship",      120, 2, 2, 4, 2, 10, 0, 8, 4, 20, 0,  8),
		new Ship("CV", "Assault Carrier",  75, 2, 1, 2, 0,  6, 0, 0, 1,  8, 6, 10),
	};

	//-----------------------------------------------------------------
	//  Basic weapon stats
	//-----------------------------------------------------------------
	Weapon[] basicWeapons = {
		new Weapon("LB", "Laser Battery",   9, 50, 1, 0, 0, true, false, false, false),
		new Weapon("LC", "Laser Cannon",   10, 60, 2, 0, 0, true, true, false, false),
		new Weapon("RB", "Rocket Battery",  3, 40, 2, 0, 3, false, false, true, false),
		new Weapon("AR", "Assault Rockets", 4, 60, 2, 4, 5, false, true, true, true),
		new Weapon("TS", "Torpedoes",       4, 50, 4, 0, 10, false, false, true, true),
	};

	//-----------------------------------------------------------------
	//  Combination point-buy squadron (6 points)
	//-----------------------------------------------------------------
	String[][] comboSquad6 = {
		new String[]{"CA"},
		new String[]{"DD", "VS"},
		new String[]{"FF", "FF"},
		new String[]{"VS", "VS", "VS"},	
	};

	//-----------------------------------------------------------------
	//  Combination point-buy squadron (12 points)
	//-----------------------------------------------------------------
	String[][] comboSquad12 = {
		new String[]{"CV", "VS"},
		new String[]{"BB", "DD"},
		new String[]{"BB", "VS", "VS"},
		new String[]{"CA", "CA"},
		new String[]{"CA", "DD", "VS"},
		new String[]{"CA", "FF", "FF"},
		new String[]{"CA", "VS", "VS", "VS"},
		new String[]{"CL", "CL", "VS"},
		new String[]{"CL", "DD", "FF"},
		new String[]{"CL", "FF", "VS", "VS"},
		new String[]{"DD", "DD", "DD"},
		new String[]{"DD", "DD", "VS", "VS"},
		new String[]{"DD", "FF", "FF", "VS"},
		new String[]{"DD", "VS", "VS", "VS", "VS"},
		new String[]{"FF", "FF", "FF", "FF"},
		new String[]{"FF", "FF", "VS", "VS", "VS"},
		new String[]{"VS", "VS", "VS", "VS", "VS", "VS"},
	};

	//-----------------------------------------------------------------
	//  Global combat records
	//-----------------------------------------------------------------
	int distance;

	//-----------------------------------------------------------------
	//  Main method
	//-----------------------------------------------------------------
	public static void main (String[] args) {
		BookOfSFKH sim = new BookOfSFKH();
		switch (simSelect) {
			case 0: sim.uniformAssess(); break;
			case 1: sim.uniformDetail(); break;
			case 2: sim.comboAssess(); break;
			case 3: sim.comboDetail(); break;
		}
	}

	//-----------------------------------------------------------------
	//  Battle two specified unit types with detail reports
	//-----------------------------------------------------------------
	void uniformDetail () {
		Ship ship1 = getShipTypeFromCode(uniDetailType1);
		Ship ship2 = getShipTypeFromCode(uniDetailType2);
		oneUniformGame(ship1, ship2);
	}

	//-----------------------------------------------------------------
	//  Create table of assessed win percents
	//    (Random budget spent on one kind of ship)
	//-----------------------------------------------------------------
	void uniformAssess () {
		makeUniformAssessTable(basicShipType, basicShipType);
	}

	//-----------------------------------------------------------------
	//  Make general assessment table
	//    (Random budget spent on one kind of ship)
	//-----------------------------------------------------------------
	void makeUniformAssessTable (Ship[] array1, Ship[] array2) {
  
		// Title
		printf("Assessed win percents "
			+ "(budget " + uniBudgetMin + "-" + uniBudgetMax + "):\n\n");

	 	// Header
		for (int i = 0; i < array2.length; i++) {
			printf("\t" + array2[i].getCode());
		}
		printf("\tWins");
		printf("\tAvgPct\n");
		for (int i = 0; i < array2.length + 3; i++) {
			printf("--------");
		}
		printf("\n");

  		// Body
		for (int i = 0; i < array1.length; i++) {
			int winCount = 0;
			int sumPctWon = 0;
			printf(array1[i].getCode() + "\t");
			for (int j = 0; j < array2.length; j++) {
				if (array1[i] == array2[j])
					printf("--\t");
				else {
					int pctWon = (int) 
						(uniformAssessGames(array1[i], array2[j]) * 100 + 0.5);
					printf(pctWon >= 50 ? pctWon + "\t" : "--\t");
					if (pctWon >= 50) winCount++;
					sumPctWon += pctWon;
				}
			}
			printf(winCount + "\t");
			printf(sumPctWon/(array1.length-1) + "\n");
		}

		// Tail
		printf("\n");
	}

	//-----------------------------------------------------------------
	//  Run many game trials; report rate of wins by first unit
	//-----------------------------------------------------------------
	double uniformAssessGames(Ship ship1, Ship ship2) {
		int wins = 0;
		for (int i = 0; i < trialsPerMatchup; i++) {
			if (oneUniformGame(ship1, ship2)) wins++;
		}
		return (double) wins/trialsPerMatchup;
	}

	//-----------------------------------------------------------------
	//  One uniform budget-based game (return: did ship1 type win?)
	//-----------------------------------------------------------------
	boolean oneUniformGame (Ship ship1, Ship ship2) {
		int budget = randomBudget();
		reportDetail("Budget: " + budget);
		Squad squad1 = squadBuyUniform(ship1, budget);
		Squad squad2 = squadBuyUniform(ship2, budget);
		return oneGame(squad1, squad2);
	}

	//-----------------------------------------------------------------
	//  Random budget
	//-----------------------------------------------------------------
	int randomBudget () {
		int range = uniBudgetMax - uniBudgetMin;
		return uniBudgetMin + (int) (Math.random() * range);
	}

	//-----------------------------------------------------------------
	//  Squad of one ship type by budget
	//-----------------------------------------------------------------
	Squad squadBuyUniform (Ship shipType, int budget) {

		// Main ships
		int shipCost = shipType.getCost();
		int numShips = budget/shipCost;
		Squad squad = new Squad(numShips, shipType);

		// Partial ship from remainder
		int remainder = budget%shipCost;
		if (partialShipFromRemainder && remainder > 0) {
			double ratio = (double) remainder/shipCost;
			squad.addNewShips(new Ship(shipType, ratio), 1);
		}
		return squad;
	}

	//-----------------------------------------------------------------
	//  Battle two specified combination squads with detail reports
	//-----------------------------------------------------------------
	void comboDetail () {
		String[] combo1 = comboSquad12[comboDetailSquad1];
		String[] combo2 = comboSquad12[comboDetailSquad2];
		oneComboGame(combo1, combo2);
	}

	//-----------------------------------------------------------------
	//  Create table of assessed win percents
	//    (Premade combination squads)
	//-----------------------------------------------------------------
	void comboAssess () {
		makeComboAssessTable(comboSquad12, comboSquad12);
	}

	//-----------------------------------------------------------------
	//  Make general assessment table
	//    (Premade combination squads)
	//-----------------------------------------------------------------
	void makeComboAssessTable (String[][] array1, String[][] array2) {
  
		// Title
		printf("Assessed win percents (combo squads):\n\n");

	 	// Header
		for (int i = 0; i < array2.length; i++) {
			printf("\t" + i);
		}
		printf("\tWins");
		printf("\tAvgPct\n");
		for (int i = 0; i < array2.length + 3; i++) {
			printf("--------");
		}
		printf("\n");

  		// Body
		for (int i = 0; i < array1.length; i++) {
			int winCount = 0;
			int sumPctWon = 0;
			printf(i + "\t");
			for (int j = 0; j < array2.length; j++) {
				if (array1[i] == array2[j])
					printf("--\t");
				else {
					int pctWon = (int) 
						(comboAssessGames(array1[i], array2[j]) * 100 + 0.5);
					printf(pctWon >= 50 ? pctWon + "\t" : "--\t");
					if (pctWon >= 50) winCount++;
					sumPctWon += pctWon;
				}
			}
			printf(winCount + "\t");
			printf(sumPctWon/(array1.length-1) + "\n");
		}

		// Tail
		printf("\n");
	}

	//-----------------------------------------------------------------
	//  Run many game trials; report rate of wins by first unit
	//-----------------------------------------------------------------
	double comboAssessGames(String[] list1, String[] list2) {
		int wins = 0;
		for (int i = 0; i < trialsPerMatchup; i++) {
			if (oneComboGame(list1, list2)) wins++;
		}
		return (double) wins/trialsPerMatchup;
	}

	//-----------------------------------------------------------------
	//  One combo-squad game (return: did list1 group win?)
	//-----------------------------------------------------------------
	boolean oneComboGame (String[] list1, String[] list2) {
		Squad squad1 = squadFromCodeList(list1);
		Squad squad2 = squadFromCodeList(list2);
		return oneGame(squad1, squad2);
	}

	//-----------------------------------------------------------------
	//   Create squad from string code list
	//-----------------------------------------------------------------
	Squad squadFromCodeList (String[] codeList) {
		Squad squad = new Squad();
		for (int i = 0; i < codeList.length; i++) {
			squad.addNewShips(getShipTypeFromCode(codeList[i]), 1);
		}
		return squad;
	}

	//-----------------------------------------------------------------
	//  One game (return: did squad profile1 win?)
	//-----------------------------------------------------------------
	boolean oneGame (Squad profile1, Squad profile2) {

		// Set up the combat
		Squad squad1 = new Squad(profile1);
		Squad squad2 = new Squad(profile2);
		initBattlefield(squad1, squad2);
		reportDetail("Setup " + squad1);
		reportDetail("Setup " + squad2);

		// Launch fighters
		launchFighters(squad1);
		launchFighters(squad2);
		
		// Let faster ships move first
		if (!squadOneMovesFirst(squad1, squad2))
			oneTurn(squad2, squad1);
		
		// Battle until one side is destroyed
		while (!isAnySquadDead(squad1, squad2)) {
			oneTurn(squad1, squad2);
			oneTurn(squad2, squad1);
		}

		// Report & return
		boolean squad1Won = squad2.isDead();
		reportDetail("* WINNER *: " + (squad1Won ? profile1 : profile2));
		return squad1Won;
	}

	//-----------------------------------------------------------------
	//  Initialize combat area
	//-----------------------------------------------------------------
	void initBattlefield (Squad squad1, Squad squad2) {

		// Assume distance 3 so all weapons usable
		// (Tried faster ship preferred range, but then no RB vs. VS)
		distance = 3;
		reportDetail("Distance: " + distance);
	}

	//-----------------------------------------------------------------
	//  Launch all fighters for a squad
	//-----------------------------------------------------------------
	void launchFighters (Squad squad) {
		for (int i = 0; i < squad.getNumShips(); i++) {
			Ship ship = squad.getShip(i);
			int numFighters = ship.getFighters();
			if (numFighters > 0) {
				reportDetail("Launching " + numFighters 
					+ " fighters from " + ship);
				squad.addNewShips(fighterShipType, numFighters);
				ship.subFighters(numFighters);
			}
		}	
	}

	//-----------------------------------------------------------------
	//  Does squad1 move first?
	//    (Compare squad speeds; if tied then flip coin)
	//-----------------------------------------------------------------
	boolean squadOneMovesFirst (Squad squad1, Squad squad2) {
		int speed1 = squadSpeed(squad1);
		int speed2 = squadSpeed(squad2);
		return (speed1 > speed2) || (speed1 == speed2 && d10() > 5);
	}

	//-----------------------------------------------------------------
	//  Squad speed factor: sum ADF+MR of lead ship
	//-----------------------------------------------------------------
	int squadSpeed (Squad squad) {
		Ship leadShip = squad.getLeadShip();
		return leadShip.getADF() + leadShip.getMR();
	}

	//-----------------------------------------------------------------
	//  One turn of action
	//-----------------------------------------------------------------
	void oneTurn (Squad squadToMove, Squad squadNoMove) {

		// SquadNoMove gets defensive fire
		if (isAnySquadDead(squadToMove, squadNoMove)) return;
		oneSquadFires(squadNoMove, squadToMove, false);
		
		// SquadToMove gets offensive fire
		if (isAnySquadDead(squadToMove, squadNoMove)) return;
		oneSquadFires(squadToMove, squadNoMove, true);
	}

	//-----------------------------------------------------------------
	//  One squad fires its weapons
	//-----------------------------------------------------------------
	void oneSquadFires (Squad shooters, Squad targets, boolean moving) {

		// Report
		reportDetail((moving ? "Offensive" : "Defensive") 
			+ " Fire by " + shooters);
		reportDetail("  Leader is " + shooters.getLeadShip());
		reportDetail("  Target is " + targets.getLeadShip());

		// All ships fire
		for (int i = 0; i < shooters.getNumShips(); i++) {
			oneShipFires(shooters.getShip(i), targets.getLeadShip(), moving);
			if (optimizeTargetSelection) {
				targets.removeDeadShips();
				if (targets.getNumShips() <= 0) break;
			}
		}
		targets.removeDeadShips();
	}

	//-----------------------------------------------------------------
	//  One ship fires its weapons
	//-----------------------------------------------------------------
	void oneShipFires (Ship shooter, Ship target, boolean moving) {
		boolean targetDeadAtStart = target.isDead();
		for (int i = 0; i < Ship.NUM_WEAPONS; i++) {
			Weapon weapon = basicWeapons[i];
			int count = shooter.getWeaponCount(i);
			if ((count > 0) 
					&& (distance <= weapon.getRange())
					&& (moving || !weapon.isMPO())
					&& (moving || !weapon.isFF() 
						|| d100() <= percentFFOnDefense)) {

				// LTD's (missiles) fire once
				if (weapon.isLTD()) {
					oneWeaponFires(weapon, target);
					shooter.decWeaponCount(i);									
				}
				
				// Non-LTD's (beams) fire each
				else {
					for (int j = 0; j < count; j++) {
						oneWeaponFires(weapon, target);
					}
				}
			}
		}	
		if (target.isDead() && !targetDeadAtStart) {
			reportDetail("  SHIP DESTROYED: " + target);		
		}
	}

	//-----------------------------------------------------------------
	//  One weapon fires
	//-----------------------------------------------------------------
	void oneWeaponFires (Weapon weapon, Ship target) {
		if (d100() <= computeToHit(weapon, target)) {
			int damage = weapon.getDmgAdd();
			for (int i = 0; i < weapon.getDmgDice(); i++) {
				damage += d10();
			}
			target.subHP(damage);
		}
	}

	//-----------------------------------------------------------------
	//  Compute to-hit for a weapon
	//-----------------------------------------------------------------
	int computeToHit (Weapon weapon, Ship target) {
		int toHit = weapon.getToHit();
		int numICMsFired = fireICMs(weapon, target);
		
		// Range diffusion
		if (weapon.isRD()) {
			toHit -= distance * 5;
		}	

		// ICMs fired vs. missiles
		if (numICMsFired > 0) {
			toHit -= numICMsFired * weapon.getModICM();
		}
	
		// VF/VS evasive maneuvers vs. torpedoes	
		if (target.getMR() >= 4 && weapon.getModICM() == 10) {
			toHit -= target.getMR() * 5;		
		}

		// Minimum is 5%
		return Math.max(toHit, 5);
	}

	//-----------------------------------------------------------------
	//  Fire ICMs vs. a weapon (return: number fired)
	//-----------------------------------------------------------------
	int fireICMs (Weapon weapon, Ship target) {
		int numFired = 0;
		if (weapon.getModICM() >= 5) {
			numFired = Math.min(ICMsUsedPerAttack, target.getICM());
			target.subICM(numFired);
		}
		return numFired;
	}

	//-----------------------------------------------------------------
	//  Is either squad destroyed?
	//-----------------------------------------------------------------
	boolean isAnySquadDead (Squad squad1, Squad squad2) {
		return squad1.isDead() || squad2.isDead();	
	}

	//-----------------------------------------------------------------
	//  Roll one 10-sided die
	//-----------------------------------------------------------------
	int d10 () {
		return (int) (Math.random() * 10) + 1;
	}

	//-----------------------------------------------------------------
	//  Roll percentile dice
	//-----------------------------------------------------------------
	int d100 () {
		return (int) (Math.random() * 100) + 1;
	}

	//-----------------------------------------------------------------
	//  Find basic ship type from code string
	//-----------------------------------------------------------------
	Ship getShipTypeFromCode (String code) {
		for (int i = 0; i < basicShipType.length; i++) {
			if (basicShipType[i].getCode().equals(code))
				return basicShipType[i];						
		}
		return null;
	}

	//-----------------------------------------------------------------
	//  Report detail for detailed game
	//-----------------------------------------------------------------
	void reportDetail (String s) {
		if (simSelect == 1 || simSelect == 3) 
			System.out.println(s);
	}
	
	//-----------------------------------------------------------------
	//  Printf recreation (for copied tables code)
	//-----------------------------------------------------------------
	void printf (String s) {
		System.out.print(s);	
	}
}


/*
=====================================================================
LICENSING INFORMATION

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published byb
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA

The author may be contacted by email at: delta@superdan.net
=====================================================================
*/

