//********************************************************************
//  Ship.java
//    Represents one spaceship.
//  Copyright (c) 2013 Daniel R. Collins. All rights reserved.
//********************************************************************

public class Ship {
	private String name, code, type;
	private int HP, ADF, MR;
	private int[] weaponCount;
	private int MS, ICM;
	private int fighters;
	private int cost;
	static public final int NUM_WEAPONS = 5;

	//-----------------------------------------------------------------
	//  Basic constructor 
	//-----------------------------------------------------------------
	public Ship (String code, String type, int HP, int ADF, int MR, 
			int weap0, int weap1, int weap2, int weap3, int weap4, 
			int MS, int ICM, int fighters, int cost) {
	
		// Parameters
		this.name = "";
		this.code = code;
		this.type = type;
		this.ADF = ADF;
		this.MR = MR;
		this.HP = HP;
		this.MS = MS;
		this.ICM = ICM;
		this.fighters = fighters;
		this.cost = cost;
		
		// Weapons
		weaponCount = new int[NUM_WEAPONS];
		weaponCount[0] = weap0;
		weaponCount[1] = weap1;
		weaponCount[2] = weap2;
		weaponCount[3] = weap3;
		weaponCount[4] = weap4;
	}

	//-----------------------------------------------------------------
	//  Copy constructor
	//-----------------------------------------------------------------
	public Ship (Ship original) {

		// Parameters
		this.name = original.name;
		this.code = original.code;
		this.type = original.type;
		this.ADF = original.ADF;
		this.MR = original.MR;
		this.HP = original.HP;
		this.MS = original.MS;
		this.ICM = original.ICM;
		this.fighters = original.fighters;
		this.cost = original.cost;
		
		// Weapons
		weaponCount = new int[NUM_WEAPONS];
		for (int i = 0; i < NUM_WEAPONS; i++) {
			this.weaponCount[i] = original.weaponCount[i];
		}
	}
	
	//-----------------------------------------------------------------
	//  Partial copy constructor (creates damaged ship)
	//-----------------------------------------------------------------
	public Ship (Ship original, double ratio) {

		// Parameters
		this.name = original.name;
		this.code = original.code;
		this.type = original.type;
		this.ADF = original.ADF;
		this.MR = original.MR;
		this.HP = (int)(original.HP * ratio);
		this.MS = (int)(original.MS * ratio);
		this.ICM = (int)(original.ICM * ratio);
		this.fighters = (int)(original.fighters * ratio);
		this.cost = (int)(original.cost * ratio);
		
		// Weapons
		weaponCount = new int[NUM_WEAPONS];
		for (int i = 0; i < NUM_WEAPONS; i++) {
			this.weaponCount[i] = (int)(original.weaponCount[i] * ratio);
		}
	}
	
	//-----------------------------------------------------------------
	//  Accessors
	//-----------------------------------------------------------------
	public String getName() { return name; };
	public String getCode() { return code; };
	public String getType() { return type; };
	public int getHP() { return HP; };
	public int getADF() { return ADF; };
	public int getMR() { return MR; };
	public int getMS() { return MS; };
	public int getICM() { return ICM; };
	public int getFighters() { return fighters; };
	public int getCost() { return cost; };
	public int getWeaponCount (int idx) { return weaponCount[idx]; };
	public boolean isDead() { return HP <= 0; };
	
	//-----------------------------------------------------------------
	//  Use resources
	//-----------------------------------------------------------------
	public void subHP(int num) { HP -= num; };
	public void subICM(int num) { ICM -= num; };
	public void subFighters(int num) { fighters -= num; };
	public void decWeaponCount(int idx) { weaponCount[idx]--; };

	//-----------------------------------------------------------------
	//  String description
	//-----------------------------------------------------------------
	public String toString() {
		return type + " (HP "+HP + ", RB "+weaponCount[2]
			+ ", AR "+weaponCount[3] + ", TS "+weaponCount[4]
			+ ", ICM "+ICM + ")";
	}
}

