//********************************************************************
//  Squad.java
//    Represents a squadron of ships.
//  Copyright (c) 2013 Daniel R. Collins. All rights reserved.
//********************************************************************

import java.util.ArrayList;

public class Squad {
	private ArrayList<Ship> ships;

	//-----------------------------------------------------------------
	//  Empty constructor
	//-----------------------------------------------------------------
	public Squad () {
		ships = new ArrayList<Ship>();
	}

	//-----------------------------------------------------------------
	//  Basic constructor
	//-----------------------------------------------------------------
	public Squad (int numShips, Ship shipType) {
		ships = new ArrayList<Ship>();
		for (int i = 0; i < numShips; i++) {
			ships.add(new Ship(shipType));		
		}
	}

	//-----------------------------------------------------------------
	//  Copy constructor
	//-----------------------------------------------------------------
	public Squad (Squad original) {
		ships = new ArrayList<Ship>();
		for (int i = 0; i < original.getNumShips(); i++) {
			ships.add(new Ship(original.getShip(i)));
		}
	}

	//-----------------------------------------------------------------
	//  Accessors
	//-----------------------------------------------------------------
	public int getNumShips() { return ships.size(); };
	public Ship getShip(int idx) { return ships.get(idx); };
	public boolean isDead() { return getNumShips() <= 0; };

	//-----------------------------------------------------------------
	//  Add new ships to the end
	//-----------------------------------------------------------------
	public void addNewShips (Ship shipType, int numShips) {
		for (int i = 0; i < numShips; i++) {
			ships.add(new Ship(shipType));		
		}
	}

	//-----------------------------------------------------------------
	//  Remove any dead ships
	//-----------------------------------------------------------------
	public void removeDeadShips () {
		for (int i = getNumShips()-1; i >= 0; i--) {
			if (getShip(i).isDead())
				ships.remove(i);		
		}
	}

	//-----------------------------------------------------------------
	//  Acquire the lead ship (end of array)
	//-----------------------------------------------------------------
	public Ship getLeadShip() {
		if (getNumShips() > 0)
			return getShip(getNumShips()-1);
		else
			return null;
	}

	//-----------------------------------------------------------------
	//  String description
	//-----------------------------------------------------------------
	public String toString() {
		if (isDead())
			return "Destroyed squad";
		else
			return getLeadShip().getCode() + "-led squad, " 
				+ getNumShips() + " ship(s)";
	}
}

